#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#define PWNDBG_COMPAT 1
#include "malloc-2.7.2.c"

#ifndef PORT
#define PORT 10000
#endif

#define KEYSZ	    8
#define MAXENTRIES 16
#define SECRETSZ   32
#define USERSZ	    8

struct entry {
	char key[KEYSZ];
	char *value;
	int  sz;
};

char secret[SECRETSZ];
char user[USERSZ];

int readsecret();
struct entry *findempty();
struct entry *search(const char *key);
int readsz();
int readkey(char *p);
int readn(char *p, int n);
int writen(char *p, int n);

struct entry entries[MAXENTRIES];

void child()
{
	char searchkey[KEYSZ + 1];
	char cmd;
	int i, sz;
	struct entry *e;

	readsecret();
	strcpy(user, "nobody");

	while (read(0, &cmd, 1) > 0) {
		switch (cmd) {
		case '\n':
			break;
		case 'n': // new
			e = findempty(entries);
			if (e == NULL)
				break;
			memset(e, 0, sizeof(struct entry));
			sz = readsz();
			if (sz <= 0)
				return;
			e->value = malloc(sz);
			if (e->value == NULL)
				return;
			if (readkey(searchkey) < 0)
				return;
			strcpy(e->key, searchkey);
			if (readn(e->value, sz) < 0)
				return;
			e->sz = sz;
			break;
		case 's': // search
			if (readkey(searchkey) < 0)
				return;
			e = search(searchkey);
			if (e == NULL)
				break;
			if (writen(e->value, e->sz) < 0)
				return;
			break;
		case 'd': // delete
			if (readkey(searchkey) < 0)
				return;
			e = search(searchkey);
			if (e == NULL)
				break;
			free(e->value);
			memset(e->key, 0, KEYSZ);
			e->value = NULL;
			e->sz = 0;
			break;
		case 'S':
			if (!strcmp(user, "root"))
				writen(secret, SECRETSZ);
			break;
		}
	}
}

int readkey(char *p)
{
	int i;

	for (i = 0; i < KEYSZ; i++) {
		if (read(0, p, 1) <= 0)
			return -1;
		if (*p == '\n')
			break;
		p++;
	}
	*p = '\0';
	return 0;
}

/* NO INTENTIONAL ERRORS STARTING FROM THIS POINT */
int readsecret()
{
	int fd, i;
	char c, *dst;

	fd = open("flag.txt", O_RDONLY);
	if (fd < 0)
		return -1;

	for (i = 0, dst = secret; i < SECRETSZ; i++, dst++) {
		if (read(fd, dst, 1) < 0)
			return -1;
	}
	close(fd);
}

struct entry *search(const char *key)
{
	struct entry *e;

	for (e = entries; e != entries + MAXENTRIES; e++) {
		if (strncmp(e->key, key, KEYSZ) == 0 && e->value != NULL)
			return e;
	}
	return NULL;
}

struct entry *findempty()
{
	struct entry *e;

	for (e = entries; e != entries + MAXENTRIES; e++) {
		if (e->value == NULL)
			return e;
	}
	return NULL;
}

int readsz()
{
	int rv;
	char c;

	rv = 0;
	while (read(0, &c, 1) >= 0) {
		if (c < '0' || c > '9')
			break;
		rv = rv * 10 + (c - '0');
	}
	return rv;
}

int readn(char *p, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (read(0, p, 1) <= 0)
			return -1;
		p++;
	}
	return 0;
}

int writen(char *p, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (write(1, p, 1) != 1)
			return -1;
		p++;
	}
	return 0;
}


void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;
	char * volatile dummy;

	srand(time(NULL));

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
